@testset "BSplineManifold" begin
    function arrayofvector2array(a::AbstractArray{<:AbstractVector{T},d})::Array{T,d+1} where {d, T<:Real}
        d̂ = length(a[1])
        s = size(a)
        N = prod(s)
        a_2dim = [a[i][j] for i in 1:N, j in 1:d̂]
        a′ = reshape(a_2dim, s..., d̂)
        return a′
    end

    function array2arrayofvector(a::AbstractArray{T,d})::Array{Vector{T},d-1} where {d, T<:Real}
        s = size(a)
        d̂ = s[end]
        N = s[1:end-1]
        a_flat = reshape(a,prod(N),d̂)
        a_vec = [a_flat[i,:] for i in 1:prod(N)]
        a′ = reshape(a_vec,N)
        return a′
    end

    @testset "constructor" begin
        k1 = KnotVector([0,0,1,1])
        k2 = UniformKnotVector(1:8)
        k3 = KnotVector(rand(8))
        _k1 = KnotVector([0,0,1,1])
        _k2 = KnotVector(1:8)
        _k3 = copy(k3)
        P1 = BSplineSpace{1}(k1)
        P2 = BSplineSpace{3}(k2)
        P3 = BSplineSpace{2}(k3)
        _P1 = BSplineSpace{1}(_k1)
        _P2 = BSplineSpace{3}(_k2)
        _P3 = BSplineSpace{2}(_k3)

        # 0-dim
        a = fill(1.2)
        @test BSplineManifold(a) == BSplineManifold(a)
        @test BSplineManifold(a) == BSplineManifold(copy(a))
        @test hash(BSplineManifold(a)) == hash(BSplineManifold(a))
        @test hash(BSplineManifold(a)) == hash(BSplineManifold(copy(a)))

        # 4-dim
        a = rand(dim(P1), dim(P2), dim(P3), dim(P3))
        @test BSplineManifold(a, P1, P2, P3, P3) == BSplineManifold(a, P1, P2, P3, P3)
        @test hash(BSplineManifold(a, P1, P2, P3, P3)) == hash(BSplineManifold(a, P1, P2, P3, P3))
        @test BSplineManifold(a, P1, P2, P3, P3) == BSplineManifold(a, _P1, _P2, _P3, _P3)
        @test hash(BSplineManifold(a, P1, P2, P3, P3)) == hash(BSplineManifold(a, _P1, _P2, _P3, _P3))
        @test BSplineManifold(a, P1, P2, P3, P3) == BSplineManifold(copy(a), _P1, _P2, _P3, _P3)
        @test hash(BSplineManifold(a, P1, P2, P3, P3)) == hash(BSplineManifold(copy(a), _P1, _P2, _P3, _P3))
    end

    @testset "1dim" begin
        @testset "BSplineManifold-1dim" begin
            P1 = BSplineSpace{1}(KnotVector([0, 0, 1, 1]))
            n1 = dim(P1) # 2
            a = [Point(i, rand()) for i in 1:n1]  # n1 × n2 array of d̂-dim vector.
            M = BSplineManifold(a, (P1,))
            @test dim(M) == 1

            P1′ = BSplineSpace{2}(KnotVector([-2, 0, 0, 1, 1, 2]))
            p₊ = (Val(1),)
            k₊ = (KnotVector(Float64[]),)

            @test P1 ⊑ P1′

            M′ = refinement(M, P1′)
            M′′ = refinement(M, p₊, k₊)
            ts = [[rand()] for _ in 1:10]
            for t in ts
                @test M(t...) ≈ M′(t...)
                @test M(t...) ≈ M′′(t...)
            end
            @test_throws DomainError M(-5)

            @test map.(M,[0.2,0.5,0.6]) == M.([0.2,0.5,0.6])
            @test M(:) == M
            @test Base.mightalias(controlpoints(M), controlpoints(M))
            @test !Base.mightalias(controlpoints(M), controlpoints(M(:)))
        end
    end

    @testset "2dim" begin
        @testset "BSplineManifold-2dim" begin
            P1 = BSplineSpace{1}(KnotVector([0, 0, 1, 1]))
            P2 = BSplineSpace{1}(KnotVector([1, 1, 2, 3, 3]))
            n1 = dim(P1) # 2
            n2 = dim(P2) # 3
            a = [Point(i, j) for i in 1:n1, j in 1:n2]  # n1 × n2 array of d̂-dim vector.
            M = BSplineManifold(a, (P1, P2))
            @test dim(M) == 2

            P1′ = BSplineSpace{2}(KnotVector([0, 0, 0, 1, 1, 1]))
            P2′ = BSplineSpace{1}(KnotVector([1, 1, 2, 1.45, 3, 3]))
            p₊ = (Val(1), Val(0))
            k₊ = (KnotVector(Float64[]), KnotVector([1.45]))

            @test P1 ⊆ P1′
            @test P2 ⊆ P2′

            M′ = refinement(M, (P1′, P2′))
            M′′ = refinement(M, p₊, k₊)
            ts = [[rand(), 1 + 2 * rand()] for _ in 1:10]
            for t in ts
                t1, t2 = t
                @test M(t1,t2) ≈ M′(t1,t2)
                @test M(t1,t2) ≈ M′′(t1,t2)
                @test M(t1,t2) ≈ M(t1,:)(t2)
                @test M(t1,t2) ≈ M(:,t2)(t1)
            end
            @test_throws DomainError M(-5,-8)

            @test M(:,:) == M
            @test Base.mightalias(controlpoints(M), controlpoints(M))
            @test !Base.mightalias(controlpoints(M), controlpoints(M(:,:)))
        end
    end

    @testset "3dim" begin
        @testset "BSplineManifold-3dim" begin
            P1 = BSplineSpace{1}(KnotVector([0, 0, 1, 1]))
            P2 = BSplineSpace{1}(KnotVector([1, 1, 2, 3, 3]))
            P3 = BSplineSpace{2}(KnotVector(rand(16)))
            n1 = dim(P1) # 2
            n2 = dim(P2) # 3
            n3 = dim(P3)
            a = [Point(i1, i2, 3) for i1 in 1:n1, i2 in 1:n2, i3 in 1:n3]
            M = BSplineManifold(a, (P1, P2, P3))
            @test dim(M) == 3

            p₊ = (Val(1), Val(0), Val(1))
            k₊ = (KnotVector(Float64[]), KnotVector([1.45]), EmptyKnotVector())

            M′′ = refinement(M, p₊, k₊)
            ts = [[rand(), 1 + 2 * rand(), 0.5] for _ in 1:10]
            for t in ts
                t1, t2, t3 = t
                @test M(t1,t2,t3) ≈ M′′(t1,t2,t3)
                @test M(t1,t2,t3) ≈ M(t1,:,:)(t2,t3)
                @test M(t1,t2,t3) ≈ M(:,t2,:)(t1,t3)
                @test M(t1,t2,t3) ≈ M(:,:,t3)(t1,t2)
                @test M(t1,t2,t3) ≈ M(t1,t2,:)(t3)
                @test M(t1,t2,t3) ≈ M(t1,:,t3)(t2)
                @test M(t1,t2,t3) ≈ M(:,t2,t3)(t1)
            end
            @test_throws DomainError M(-5,-8,-120)

            @test M(:,:,:) == M
            @test Base.mightalias(controlpoints(M), controlpoints(M))
            @test !Base.mightalias(controlpoints(M), controlpoints(M(:,:,:)))
        end
    end
end
